//***********************************************************************//
//																		 //
//		- "Talk to me like I'm a 3 year old!" Programming Lessons -		 //
//                                                                       //
//		$Author:		DigiBen		digiben@gametutorials.com			 //
//																		 //
//		$Program:		ASE Loader										 //
//																		 //
//		$Description:	Demonstrates how to load a .Ase file format		 //
//																		 //
//		$Date:			10/13/01										 //
//																		 //
//***********************************************************************//

#include "unicycle.h"

    // This holds the 3D Model info that we load in
   bool ase::ImportASE(t3DModel *pModel, char *strFileName)
{
	char strMessage[255] = {0};				// This will be used for error messages

	// Make sure we have a valid model and file name
	if(!pModel || !strFileName) return false;

	// Here we open the desired file for read only and return the file pointer
	m_FilePointer = fopen(strFileName, "r");

	// Check to make sure we have a valid file pointer
	if(!m_FilePointer) {
		// Create an error message for the attempted file
		sprintf(strMessage, "Unable to find or open the file: %s", strFileName);
		MessageBox(NULL, strMessage, "Error", MB_OK);
		return false;
	}

	// Now that we have a valid file and it's open, let's read in the info!
	ReadAseFile(pModel);

	// Now that we have the file read in, let's compute the vertex normals for lighting
	ComputeNormals(pModel);

	// Close the .ase file that we opened
	fclose(m_FilePointer);

	// Return a success!
	return true;
}


/////	This function reads the data for every object and it's associated material
void ase::ReadAseFile(t3DModel *pModel)
{
	tMaterialInfo newMaterial = {0};		// This will be used to push on a new material
	t3DObject     newObject   = {0};		// This will be used to push on a new object
	int i = 0;	

	// This will return the number of objects stored in the .ase file
	pModel->numOfObjects   = GetObjectCount();

	// This will return the number of materials stored in the .ase file
	pModel->numOfMaterials = GetMaterialCount();

	// Go through all the materials and fill in their data and info
	for(i = 0; i < pModel->numOfMaterials; i++)
	{
		// Add a new material to our list of materials using the STL "vector" class
		pModel->pMaterials.push_back(newMaterial);

		// Get the material info for the current material.  We add 1 because we
		// want to start at 1 and i starts at 0.
		GetTextureInfo(&(pModel->pMaterials[i]), i + 1);
	}

	// Go through all the objects and fill in their data and info
	for(i = 0; i < pModel->numOfObjects; i++)
	{	
		// Add a new object to our list of objects using the STL "vector" class
		pModel->pObject.push_back(newObject);

		// Set the material ID to -1 to initialize it.  This will be changed
		// if there is a texture/material assigned to this object.
		pModel->pObject[i].materialID = -1;

		// Move the file pointer to the desired object.  We add one because our
		// object count starts at 1 and i starts at 0
		MoveToObject(i + 1);

		// Find out the number of vertices, faces and texture coordinates for this object, 
		// then allocate the memory needed to store that amount that needs to be read in.
		ReadObjectInfo(&(pModel->pObject[i]), i + 1);
		
		// Read the vertices, face indices and texture coordinates for this object
		ReadObjectData(pModel, &(pModel->pObject[i]), i + 1);
	}
}




/////	This function returns the total object count in the .ase file
int ase::GetObjectCount()
{
	char strWord[255] = {0};
	int objectCount = 0;

	// This function goes through the file and reads each word.  When it
	// finds an OBJECT tag, it increases the object count.

	// Go to the beginning of the file.  Then we can start fresh and get a good count.
	rewind(m_FilePointer);

	// Go through the whole file and end when we reached the END
	while (!feof(m_FilePointer))
	{
		// Get each word in the file
		fscanf(m_FilePointer, "%s", &strWord);

		// Check if we hit the start of an object
		if (!strcmp(strWord, OBJECT))
		{
			// Increase the current object we are at 
			objectCount++;
		}
		else
		{
			// Go to the next line
			fgets(strWord, 100, m_FilePointer);
		}
	}
	// Return the object count in the .ase file
	return objectCount;
}




/////	This function returns the total material count for the .ase file
int ase::GetMaterialCount()
{
	char strWord[255] = {0};
	int materialCount = 0;

	// This function goes through the file and reads each word.  When it
	// finds the MATERIAL_COUNT tag, it reads the material count.

	// Go to the beginning of the file
	rewind(m_FilePointer);

	// GO through the whole file until we hit the end
	while (!feof(m_FilePointer))
	{
		// Get each word in the file
		fscanf(m_FilePointer, "%s", &strWord);

		// Check if we hit the start of an object
		if (!strcmp(strWord, MATERIAL_COUNT))
		{
			// Read in the material count
			fscanf(m_FilePointer, "%d", &materialCount);

			// Return the material count
			return materialCount;
		}
		else
		{
			// Go to the next line
			fgets(strWord, 100, m_FilePointer);
		}
	}

	// Return NO materials if we get here
	return 0;
}




/////	This function reads in the information about the desired material  //////
void ase::GetTextureInfo (tMaterialInfo *pTexture, int desiredMaterial)
{
	char strWord[255] = {0};
	int materialCount= 0;
	
	// In this function we check for an MATERIAL tag, then see if it's the
	// material we want by the desiredMaterial number.  If so we start reading it in.

	// Go to the beginning of the file
	rewind(m_FilePointer);

	// Go through the whole file until we reach the end
	while (!feof(m_FilePointer))
	{
		// Get each word from the file
		fscanf(m_FilePointer, "%s", &strWord);

		// Check if we hit the start of an object
		if (!strcmp(strWord, MATERIAL))
		{
			// Increase the current object we are at 
			materialCount++;

			// Check if it's the one we want to stop at, if so break
			if(materialCount == desiredMaterial)
				break;
		}
		else
		{
			// Go to the next line
			fgets(strWord, 100, m_FilePointer);
		}
	}

	// Now we are at the material we want, so let's read it's data in

	// Go through the rest of the file until we hit the end
	while (!feof(m_FilePointer))
	{
		// Get each word from the file
		fscanf(m_FilePointer, "%s", &strWord);

		// If we found a MATERIAL tag stop because we went to far
		if (!strcmp (strWord, MATERIAL))
		{
			// We hit a new material, so we need to stop
			return;
		}
		// If we hit a MATERIAL_COLOR tag, we need to get the material's color
		else if (!strcmp(strWord, MATERIAL_COLOR))
		{
			// Get the material RGB color of the object
			fscanf(m_FilePointer, " %f %f %f", &(pTexture->fColor[0]), 
											   &(pTexture->fColor[1]), 
											   &(pTexture->fColor[2]));
		}
		// If we hit a TEXTURE tag, we need to get the texture's name
		else if (!strcmp(strWord, TEXTURE))
		{
			// Get the file name of the texture
			GetTextureName(pTexture);
		}
		// If we hit a MATERIAL_NAME tag, we need to get the material's name
		else if (!strcmp(strWord, MATERIAL_NAME))
		{
			// Get the material name of the object
			GetMaterialName(pTexture);
		}
		// If we hit a UTILE tag, we need to get the U tile ratio
		else if(!strcmp(strWord, UTILE))
		{
			// Read the U tiling for the U coordinates of the texture
			pTexture->uTile = ReadFloat();
		}
		// If we hit a VTILE tag, we need to get the V tile ratio
		else if(!strcmp(strWord, VTILE))
		{
			// Read the V tiling for the V coordinates of the texture
			pTexture->vTile = ReadFloat();
		}
		// Otherwise ignore the data and read past it
		else
		{
			// Go to the next line
			fgets (strWord, 100, m_FilePointer);
		}
	}
}



/////	This function moves the file pointer in the .ase file to the desired object  //////

void ase::MoveToObject (int desiredObject)
{
	char strWord[255] = {0};
	int objectCount = 0;

	// This function takes the number of the desired object to move to.

	// Go to the beginning of the file
	rewind(m_FilePointer);

	// Go through the whole file until we reach the end
	while(!feof(m_FilePointer))
	{
		// Get each word
		fscanf(m_FilePointer, "%s", &strWord);

		// Check if we hit the start of an object
		if(!strcmp(strWord, OBJECT))
		{
			// Increase the current object we are at 
			objectCount++;

			// Check if it's the one we want to stop at, if so stop reading
			if(objectCount == desiredObject)
				return;
		}
		else
		{
			// Go to the next line and skip this current line
			fgets(strWord, 100, m_FilePointer);
		}
	}
}

/////	This function reads in and returns a float from the .ase file     //////

float ase::ReadFloat()
{
	float v = 0.0f;

	// Read in a float
	fscanf(m_FilePointer, " %f", &v);

	// Return the float
	return v;
}


/////	This function reads in the information about the object, but not the data     /////7

void ase::ReadObjectInfo(t3DObject *pObject, int desiredObject)
{
	char strWord[255] = {0};

	// In this function we position the file pointer to the object we want
	// to read in, then we read until we find the associated tags, then read the info.

	// Go to the desired object to read from in the file
	MoveToObject(desiredObject);
	
	// While we are not at the end of the file
	while (!feof(m_FilePointer))
	{
		// Read in each word from the file to check against
		fscanf(m_FilePointer, "%s", &strWord);

		// If we hit the number of vertices tag
		if (!strcmp(strWord, NUM_VERTEX))
		{
			// Read in the number of vertices for this object
			fscanf(m_FilePointer, "%d", &pObject->numOfVerts);

			// Allocate enough memory to hold the vertices
			pObject->pVerts = new CVector3 [pObject->numOfVerts];
		}
		// If we hit the number of faces tag
		else if (!strcmp(strWord, NUM_FACES))
		{
			// Read in the number of faces for this object
			fscanf(m_FilePointer, "%d", &pObject->numOfFaces);

			// Allocate enough memory to hold the faces
			pObject->pFaces = new tFace [pObject->numOfFaces];
		}
		// If we hit the number of texture vertices tag
		else if (!strcmp(strWord, NUM_TVERTEX))
		{
			// Read in the number of texture coordinates for this object
			fscanf(m_FilePointer, "%d", &pObject->numTexVertex);

			// Allocate enough memory for the UV coordinates
			pObject->pTexVerts = new CVector2 [pObject->numTexVertex];
		}
		// If we hit the object tag we want to stop cause we went to far
		else if (!strcmp(strWord, OBJECT))	
		{
			// Return if we get to the next object
			return;
		}
		else 
		{
			// We didn't find anything we want to check for so we read to the next line
			fgets(strWord, 100, m_FilePointer);
		}
	}	
}


/////	This function reads in the file name of the texture assigned to the object //////

void ase::GetTextureName(tMaterialInfo *pTexture)
{
	// Read in the texture's file name
	fscanf (m_FilePointer, " \"%s", &(pTexture->strFile));
	
	// Put a NULL character at the end of the string
	pTexture->strFile[strlen (pTexture->strFile) - 1] = '\0';
}




/////	This function reads in the material name of the object       //////

void ase::GetMaterialName(tMaterialInfo *pTexture)
{
	// Read in the material's name (Make sure this is just one word or it won't work well)
	fscanf (m_FilePointer, " \"%s", &(pTexture->strName));
	
	// Put a NULL character at the end of the string just in case
	pTexture->strName[strlen (pTexture->strName)] = '\0';
}


/////	This function calls the functions that query and read in the desired data  //////

void ase::ReadObjectData(t3DModel *pModel, t3DObject *pObject, int desiredObject)
{

	// Load the material ID for this object
	GetData(pModel, pObject, MATERIAL_ID, desiredObject);

	// Load the vertices for this object
	GetData(pModel, pObject, VERTEX,		 desiredObject);

	// Load the texture coordinates for this object
	GetData(pModel, pObject, TVERTEX,	 desiredObject);

	// Load the vertex faces list for this object
	GetData(pModel, pObject, FACE,		 desiredObject);

	// Load the texture face list for this object
	GetData(pModel, pObject, TFACE,		 desiredObject);

	// Load the texture for this object
	GetData(pModel, pObject, TEXTURE,	 desiredObject);

	// Load the U tile for this object
	GetData(pModel, pObject, UTILE,		 desiredObject);

	// Load the V tile for this object
	GetData(pModel, pObject, VTILE,		 desiredObject);
}




/////	This function is the loop for reading in the object data  ////////

void ase::GetData(t3DModel *pModel, t3DObject *pObject, char *strDesiredData, int desiredObject)
{
	char strWord[255] = {0};
	
	// Move the file pointer to the desired object
	MoveToObject(desiredObject);

	// Go through the file until we reach the end
	while(!feof(m_FilePointer))
	{
		// Read in every word to check it against tags
		fscanf(m_FilePointer, "%s", &strWord);

		// If we reached an object tag, stop read because we went to far
		if(!strcmp(strWord, OBJECT))	
		{
			// Stop reading because we are done with the current object
			return;
		}
		// If we hit a vertex tag
		else if(!strcmp(strWord, VERTEX))
		{
			// Make sure that is the data that we want to read in
			if(!strcmp(strDesiredData, VERTEX)) 
			{
				// Read in a vertex
				ReadVertex(pObject);
			}
		}
		// If we hit a texture vertex
		else if(!strcmp(strWord, TVERTEX))
		{
			// Make sure that is the data that we want to read in
			if(!strcmp(strDesiredData, TVERTEX)) 
			{
				// Read in a texture vertex
				ReadTextureVertex(pObject, pModel->pMaterials[pObject->materialID]);
			}
		}
		// If we hit a vertice index to a face
		else if(!strcmp(strWord, FACE))
		{
			// Make sure that is the data that we want to read in
			if(!strcmp(strDesiredData, FACE)) 
			{
				// Read in a face
				ReadFace(pObject);
			}
		}
		// If we hit a texture index to a face
		else if(!strcmp(strWord, TFACE))
		{
			// Make sure that is the data that we want to read in
			if(!strcmp(strDesiredData, TFACE))
			{
				// Read in a texture indice for a face
				ReadTextureFace(pObject);
			}
		}
		// If we hit the material ID to the object
		else if(!strcmp(strWord, MATERIAL_ID))
		{
			// Make sure that is the data that we want to read in
			if(!strcmp(strDesiredData, MATERIAL_ID))
			{
				// Read in the material ID assigned to this object
				pObject->materialID = (int)ReadFloat();
				return;
			}				
		}
		else 
		{
			// We must not care about this tag read so read past the whole line
			fgets(strWord, 100, m_FilePointer);
		}
	}
}




/////	This function reads in the vertices for the object   ////////

void ase::ReadVertex(t3DObject *pObject)
{
	int index = 0;

	// Read past the vertex index
	fscanf(m_FilePointer, "%d", &index);
	
	// Here we read in the vertice's (X, Y, Z).  Since 3D Studio Max has Z pointing
	// up, we are going to swap the Y and Z to make it normal.  Otherwise we would need
	// to rotate it around the X axis by -90 degrees: glRotatef(-90, 1.0f, 0.0f, 0.0f);
	fscanf(m_FilePointer, "%f %f %f", &pObject->pVerts[index].x, 
									  &pObject->pVerts[index].z,
									  &pObject->pVerts[index].y);
		
	// Also in 3D studio max the Z goes out of the screen so we want to swap that too.
	// We do that by negating the Z value of each vertex.
	pObject->pVerts[index].z = -pObject->pVerts[index].z;
}

/////	This function reads in the texture coordinates  ////////

void ase::ReadTextureVertex(t3DObject *pObject, tMaterialInfo texture)
{
	int index = 0;

	// Here we read past the index of the texture coordinate
	fscanf(m_FilePointer, "%d", &index);

	// Next, we read in the (U, V) texture coordinates.
	fscanf(m_FilePointer, "%f %f", &(pObject->pTexVerts[index].x), &(pObject->pTexVerts[index].y));

	// What is being done here is we are multiplying a X and Y tile factor
	// to the UV coordinate.  Usually the uTile and vTile is 1, but if it depends on
	// your UVW map.  If you made the texture tile more along the object, these would change.
	pObject->pTexVerts[index].x *= texture.uTile;
	pObject->pTexVerts[index].y *= texture.vTile;

	// We know this object has a texture so let's set this true
	pObject->bHasTexture = true;
}


/////	This function reads in the indices to the array of vertices  //////

void ase::ReadFace(t3DObject *pObject)
{
	int index = 0;

	// Read past the index of this Face
	fscanf(m_FilePointer, "%d:", &index);

	// Now we read in the actual vertex indices; One index for each point in the triangle.
	// These indices will index into the vertex array pVerts[].
	fscanf(m_FilePointer, "\tA:\t%d B:\t%d C:\t%d", &(pObject->pFaces[index].vertIndex[0]), 
													&(pObject->pFaces[index].vertIndex[1]), 
													&(pObject->pFaces[index].vertIndex[2])); 
}


/////	This function reads in the indices to the texture coordinate array ///////

void ase::ReadTextureFace(t3DObject *pObject)
{
	int index = 0;

	// Read past the index for this texture coordinate
	fscanf(m_FilePointer, "%d:", &index);

	// Now we read in the UV coordinate index for the current face.
	// This will be an index into pTexCoords[] for each point in the face.
	fscanf(m_FilePointer, "%d %d %d", &pObject->pFaces[index].coordIndex[0], 
									  &pObject->pFaces[index].coordIndex[1], 
									  &pObject->pFaces[index].coordIndex[2]);
}


//////////////////////////////	Math Functions  ////////////////////////////////*

// This computes the magnitude of a normal.   (magnitude = sqrt(x^2 + y^2 + z^2)
#define Mag(Normal) (sqrt(Normal.x*Normal.x + Normal.y*Normal.y + Normal.z*Normal.z))

// This calculates a vector between 2 points and returns the result
CVector3 Vector(CVector3 vPoint1, CVector3 vPoint2)
{
	CVector3 vVector;							// The variable to hold the resultant vector

	vVector.x = vPoint1.x - vPoint2.x;			// Subtract point1 and point2 x's
	vVector.y = vPoint1.y - vPoint2.y;			// Subtract point1 and point2 y's
	vVector.z = vPoint1.z - vPoint2.z;			// Subtract point1 and point2 z's

	return vVector;								// Return the resultant vector
}

// This adds 2 vectors together and returns the result
CVector3 AddVector(CVector3 vVector1, CVector3 vVector2)
{
	CVector3 vResult;							// The variable to hold the resultant vector
	
	vResult.x = vVector2.x + vVector1.x;		// Add Vector1 and Vector2 x's
	vResult.y = vVector2.y + vVector1.y;		// Add Vector1 and Vector2 y's
	vResult.z = vVector2.z + vVector1.z;		// Add Vector1 and Vector2 z's

	return vResult;								// Return the resultant vector
}

// This divides a vector by a single number (scalar) and returns the result
CVector3 DivideVectorByScaler(CVector3 vVector1, float Scaler)
{
	CVector3 vResult;							// The variable to hold the resultant vector
	
	vResult.x = vVector1.x / Scaler;			// Divide Vector1's x value by the scaler
	vResult.y = vVector1.y / Scaler;			// Divide Vector1's y value by the scaler
	vResult.z = vVector1.z / Scaler;			// Divide Vector1's z value by the scaler

	return vResult;								// Return the resultant vector
}

// This returns the cross product between 2 vectors
CVector3 Cross(CVector3 vVector1, CVector3 vVector2)
{
	CVector3 vCross;								// The vector to hold the cross product
												// Get the X value
	vCross.x = ((vVector1.y * vVector2.z) - (vVector1.z * vVector2.y));
												// Get the Y value
	vCross.y = ((vVector1.z * vVector2.x) - (vVector1.x * vVector2.z));
												// Get the Z value
	vCross.z = ((vVector1.x * vVector2.y) - (vVector1.y * vVector2.x));

	return vCross;								// Return the cross product
}

// This returns the normal of a vector
CVector3 Normalize(CVector3 vNormal)
{
	double Magnitude;							// This holds the magnitude			

	Magnitude = Mag(vNormal);					// Get the magnitude

	vNormal.x /= (float)Magnitude;				// Divide the vector's X by the magnitude
	vNormal.y /= (float)Magnitude;				// Divide the vector's Y by the magnitude
	vNormal.z /= (float)Magnitude;				// Divide the vector's Z by the magnitude

	return vNormal;								// Return the normal
}

/////	This function computes the normals and vertex normals of the objects

void ase::ComputeNormals(t3DModel *pModel)
{
	CVector3 vVector1, vVector2, vNormal, vPoly[3];

	// If there are no objects, we can skip this part
	if(pModel->numOfObjects <= 0)
		return;

	// Go through each of the objects to calculate their normals
	for(int index = 0; index < pModel->numOfObjects; index++)
	{
		// Get the current object
		t3DObject *pObject = &(pModel->pObject[index]);

		// Here we allocate all the memory we need to calculate the normals
		CVector3 *pNormals		= new CVector3 [pObject->numOfFaces];
		CVector3 *pTempNormals	= new CVector3 [pObject->numOfFaces];
		pObject->pNormals		= new CVector3 [pObject->numOfVerts];

		// Go though all of the faces of this object
		for(int i=0; i < pObject->numOfFaces; i++)
		{												
			// To cut down LARGE code, we extract the 3 points of this face
			vPoly[0] = pObject->pVerts[pObject->pFaces[i].vertIndex[0]];
			vPoly[1] = pObject->pVerts[pObject->pFaces[i].vertIndex[1]];
			vPoly[2] = pObject->pVerts[pObject->pFaces[i].vertIndex[2]];

			// Now let's calculate the face normals (Get 2 vectors and find the cross product of those 2)

			vVector1 = Vector(vPoly[0], vPoly[2]);		// Get the vector of the polygon (we just need 2 sides for the normal)
			vVector2 = Vector(vPoly[2], vPoly[1]);		// Get a second vector of the polygon

			vNormal  = Cross(vVector1, vVector2);		// Return the cross product of the 2 vectors (normalize vector, but not a unit vector)
			pTempNormals[i] = vNormal;					// Save the un-normalized normal for the vertex normals
			vNormal  = Normalize(vNormal);				// Normalize the cross product to give us the polygons normal

			pNormals[i] = vNormal;						// Assign the normal to the list of normals
		}

		//////////////// Now Get The Vertex Normals /////////////////

		CVector3 vSum = {0.0, 0.0, 0.0};
		CVector3 vZero = vSum;
		int shared=0;

		for (int i = 0; i < pObject->numOfVerts; i++)			// Go through all of the vertices
		{
			for (int j = 0; j < pObject->numOfFaces; j++)	// Go through all of the triangles
			{												// Check if the vertex is shared by another face
				if (pObject->pFaces[j].vertIndex[0] == i || 
					pObject->pFaces[j].vertIndex[1] == i || 
					pObject->pFaces[j].vertIndex[2] == i)
				{
					vSum = AddVector(vSum, pTempNormals[j]);// Add the un-normalized normal of the shared face
					shared++;								// Increase the number of shared triangles
				}
			}      
			
			// Get the normal by dividing the sum by the shared.  We negate the shared so it has the normals pointing out.
			pObject->pNormals[i] = DivideVectorByScaler(vSum, float(-shared));

			// Normalize the normal for the final vertex normal
			pObject->pNormals[i] = Normalize(pObject->pNormals[i]);	

			vSum = vZero;									// Reset the sum
			shared = 0;										// Reset the shared
		}
	
		// Free our memory and start over on the next object
		delete [] pTempNormals;
		delete [] pNormals;
	}
}

